@@ -1,4 +1,4 @@
-# Copyright (C) 2003-2021, Stefan Schwarzer <sschwarzer@sschwarzer.net>
+# Copyright (C) 2003-2022, Stefan Schwarzer <sschwarzer@sschwarzer.net>
 # and ftputil contributors (see `doc/contributors.txt`)
 # See the file LICENSE for licensing terms.
 
@@ -142,7 +142,10 @@ class RealFTPTest:
 
 
 class TestMkdir(RealFTPTest):
-    def test_mkdir_rmdir(self):
+    def test_directory_exists_after_mkdir(self):
+        """
+        After a `mkdir` call, the directory should exist.
+        """
         host = self.host
         dir_name = "_testdir_"
         file_name = host.path.join(dir_name, "_nonempty_")
@@ -151,29 +154,12 @@ class TestMkdir(RealFTPTest):
         host.mkdir(dir_name)
         files = host.listdir(host.curdir)
         assert dir_name in files
-        # Try to remove a non-empty directory.
-        self.cleaner.add_file(file_name)
-        non_empty = host.open(file_name, "w")
-        non_empty.close()
-        with pytest.raises(ftputil.error.PermanentError):
-            host.rmdir(dir_name)
-        # Remove file.
-        host.unlink(file_name)
-        # `remove` on a directory should fail.
-        try:
-            try:
-                host.remove(dir_name)
-            except ftputil.error.PermanentError as exc:
-                assert str(exc).startswith("remove/unlink can only delete files")
-            else:
-                pytest.fail("we shouldn't have come here")
-        finally:
-            # Delete empty directory.
-            host.rmdir(dir_name)
-        files = host.listdir(host.curdir)
-        assert dir_name not in files
 
     def test_makedirs_without_existing_dirs(self):
+        """
+        If some directories in a `makedirs` call don't exist yet, they should
+        be implcitly created.
+        """
         host = self.host
         # No `_dir1_` yet
         assert "_dir1_" not in host.listdir(host.curdir)
@@ -216,11 +202,19 @@ class TestMkdir(RealFTPTest):
         assert not host.path.isdir("/_dir1_/_dir2_")
 
     def test_makedirs_of_existing_directory(self):
+        """
+        Calling `makedirs` on an existing directory with `exist_ok` should do
+        nothing.
+        """
         host = self.host
         # The (chrooted) login directory
         host.makedirs("/", exist_ok=True)
 
     def test_makedirs_with_file_in_the_way(self):
+        """
+        If a part of `makedirs`'s path is a file, a `PermanentError` should be
+        raised.
+        """
         host = self.host
         self.cleaner.add_dir("_dir1_")
         host.mkdir("_dir1_")
@@ -232,6 +226,10 @@ class TestMkdir(RealFTPTest):
             host.makedirs("_dir1_/file1/dir2")
 
     def test_makedirs_with_existing_directory(self):
+        """
+        An already-existent subdirectory in the path of `makedirs` should be
+        ignored.
+        """
         host = self.host
         self.cleaner.add_dir("_dir1_")
         host.mkdir("_dir1_")
@@ -241,33 +239,175 @@ class TestMkdir(RealFTPTest):
         assert host.path.isdir("_dir1_/dir2")
 
     def test_makedirs_in_non_writable_directory(self):
+        """
+        If `makedirs` is asked to create a directory in a non-writable
+        directory, a `PermanentError` should be raised.
+        """
         host = self.host
         # Preparation: `rootdir1` exists but is only writable by root.
         with pytest.raises(ftputil.error.PermanentError):
             host.makedirs("rootdir1/dir2")
 
     def test_makedirs_with_writable_directory_at_end(self):
+        """
+        If `makedirs` is asked to create a directory in a writable directory
+        under a non-writable directory, the `makedirs` call should succeed.
+        """
         host = self.host
         self.cleaner.add_dir("rootdir2/dir2")
         # Preparation: `rootdir2` exists but is only writable by root. `dir2`
         # is writable by regular ftp users. Both directories below should work.
         host.makedirs("rootdir2/dir2", exist_ok=True)
         host.makedirs("rootdir2/dir2/dir3")
+        assert host.path.isdir("rootdir2/dir2")
+        assert host.path.isdir("rootdir2/dir2/dir3")
 
 
 class TestRemoval(RealFTPTest):
-    def test_rmtree_without_error_handler(self):
+
+    # Tests for `remove`
+
+    def test_remove_existing_file(self):
+        """
+        Removal of an existing file should succeed.
+        """
+        self.cleaner.add_file("_testfile_")
+        self.make_remote_file("_testfile_")
         host = self.host
-        # Build a tree.
+        assert host.path.isfile("_testfile_")
+        host.remove("_testfile_")
+        assert not host.path.exists("_testfile_")
+
+    def test_cache_invalidation_for_remove_exception(self):
+        """
+        If a file is removed, its stats information should be removed from the
+        cache. This should also work if the removal raises an exception.
+        """
+        # Test for ticket #150
+        self.cleaner.add_file("_testfile_")
+        self.make_remote_file("_testfile_")
+        host = self.host
+        assert host.path.isfile("_testfile_")
+        # Monkey-patch session `delete` call.
+        old_delete = host._session.delete
+        def failing_delete(path):
+            # Simulate the case where the file is removed on the server, but a
+            # proper reply doesn't get through to the client. It doesn't matter
+            # whether the exception is `error_temp` or `error_perm`.
+            old_delete(path)
+            raise ftplib.error_perm("simulated error")
+
+        host._session.delete = failing_delete
+        with pytest.raises(ftputil.error.PermanentError):
+            host.remove("_testfile_")
+        assert not host.path.exists("_testfile_")
+
+    def test_remove_non_existent_item(self):
+        """
+        If trying to remove a non-existent file system item, a `PermanentError`
+        should be raised.
+        """
+        host = self.host
+        with pytest.raises(ftputil.error.PermanentError):
+            host.remove("nonexistent")
+
+    def test_remove_on_directory(self):
+        """
+        Calling `remove` on a directory should raise a `PermanentError`.
+
+        (Directories must be removed with `rmdir`.)
+        """
+        host = self.host
+        dir_name = "_testdir_"
+        self.cleaner.add_dir(dir_name)
+        host.mkdir(dir_name)
+        try:
+            try:
+                host.remove(dir_name)
+            except ftputil.error.PermanentError as exc:
+                assert str(exc).startswith("remove/unlink can only delete files")
+            else:
+                pytest.fail("we shouldn't have come here")
+        finally:
+            # Delete empty directory.
+            host.rmdir(dir_name)
+        files = host.listdir(host.curdir)
+        assert dir_name not in files
+
+    # Test for `rmdir`
+
+    def test_remove_existing_directory(self):
+        """
+        If `rmdir` is called on an empty directory, the call should succeed.
+        """
+        self.cleaner.add_dir("_testdir_")
+        host = self.host
+        host.mkdir("_testdir_")
+        assert host.path.isdir("_testdir_")
+        host.rmdir("_testdir_")
+        assert not host.path.exists("_testdir_")
+
+    def test_rmdir_on_nonempty_directory(self):
+        """
+        If a directory exists, but isn't empty, `rmdir` should raise a
+        `PermanentError`.
+        """
+        host = self.host
+        dir_name = "_testdir_"
+        self.cleaner.add_dir(dir_name)
+        host.mkdir(dir_name)
+        # Try to remove a non-empty directory.
+        file_name = host.path.join(dir_name, "_nonempty_")
+        self.cleaner.add_file(file_name)
+        non_empty = host.open(file_name, "w")
+        non_empty.close()
+        with pytest.raises(ftputil.error.PermanentError):
+            host.rmdir(dir_name)
+
+    def test_cache_invalidation_for_rmdir_exception(self):
+        """
+        If a directory is removed, its stats information should be removed from
+        the cache. This should also work if the removal raises an exception.
+        """
+        # Test for ticket #150
+        self.cleaner.add_dir("_testdir_")
+        host = self.host
+        host.mkdir("_testdir_")
+        assert host.path.isdir("_testdir_")
+        # Monkey-patch session `rmd` call.
+        old_rmd = host._session.rmd
+        def failing_rmd(path):
+            # Simulate the case where the directory is removed on the server,
+            # but a proper reply doesn't get through to the client. It doesn't
+            # matter whether the exception is `error_temp` or `error_perm`.
+            old_rmd(path)
+            raise ftplib.error_perm("simulated error")
+
+        host._session.rmd = failing_rmd
+        with pytest.raises(ftputil.error.PermanentError):
+            host.rmdir("_testdir_")
+        assert not host.path.exists("_testdir_")
+
+    # Tests for `rmtree`
+
+    def build_tree(self, host):
+        """
+        Build a directory tree for tests.
+        """
         self.cleaner.add_dir("_dir1_")
         host.makedirs("_dir1_/dir2")
         self.make_remote_file("_dir1_/file1")
         self.make_remote_file("_dir1_/file2")
         self.make_remote_file("_dir1_/dir2/file3")
         self.make_remote_file("_dir1_/dir2/file4")
-        # Try to remove a _file_ with `rmtree`.
-        with pytest.raises(ftputil.error.PermanentError):
-            host.rmtree("_dir1_/file2")
+
+    def test_rmtree_without_error_handler(self):
+        """
+        Calling `rmtree` on a directory tree should remove the directory and
+        everything under it.
+        """
+        host = self.host
+        self.build_tree(host)
         # Remove `dir2`.
         host.rmtree("_dir1_/dir2")
         assert not host.path.exists("_dir1_/dir2")
@@ -280,6 +420,9 @@ class TestRemoval(RealFTPTest):
         assert not host.path.exists("_dir1_")
 
     def test_rmtree_with_error_handler(self):
+        """
+        If an `rmtree` call specifies an error handler, it should be used.
+        """
         host = self.host
         self.cleaner.add_dir("_dir1_")
         host.mkdir("_dir1_")
@@ -307,18 +450,14 @@ class TestRemoval(RealFTPTest):
         assert log[1][0] == host.rmdir
         assert log[1][1] == "_dir1_"
 
-    def test_remove_non_existent_item(self):
+    def test_remove_file_with_rmtree(self):
+        """
+        Calling `rmtree` on a file should raise a `PermanentError`.
+        """
         host = self.host
+        self.build_tree(host)
         with pytest.raises(ftputil.error.PermanentError):
-            host.remove("nonexistent")
-
-    def test_remove_existing_file(self):
-        self.cleaner.add_file("_testfile_")
-        self.make_remote_file("_testfile_")
-        host = self.host
-        assert host.path.isfile("_testfile_")
-        host.remove("_testfile_")
-        assert not host.path.exists("_testfile_")
+            host.rmtree("_dir1_/file2")
 
 
 class TestWalk(RealFTPTest):
@@ -425,15 +564,102 @@ class TestWalk(RealFTPTest):
 
 class TestRename(RealFTPTest):
     def test_rename(self):
+        """
+        If `rename` is called on two paths, the first should disappear and the
+        second should exist afterward.
+
+        This test also tests `Path` support.
+        """
         host = self.host
-        # Make sure the target of the renaming operation is removed.
+        # Make sure both files are gone after the test.
+        self.cleaner.add_file("_testfile1_")
         self.cleaner.add_file("_testfile2_")
         self.make_remote_file("_testfile1_")
         host.rename(pathlib.Path("_testfile1_"), "_testfile2_")
         assert not host.path.exists("_testfile1_")
         assert host.path.exists(pathlib.Path("_testfile2_"))
 
+    def test_cache_invalidation_for_rename(self):
+        """
+        A `rename` call should invalidate the cache entries for both path
+        arguments. The purpose of the first invalidation is obvious, but the
+        second invalidation is also important in case the path of the target
+        existed before the rename and changes according to the new file stat
+        information.
+
+        This test also tests `Path` support.
+        """
+        # Test for ticket #150
+        host = self.host
+        # Make sure both files are gone after the test.
+        self.cleaner.add_file("_testfile1_")
+        self.cleaner.add_file("_testfile2_")
+        # Case 1: Target file doesn't exist yet.
+        self.make_remote_file("_testfile1_")
+        file1_stat = host.stat("_testfile1_")
+        host.rename(pathlib.Path("_testfile1_"), "_testfile2_")
+        assert not host.path.exists("_testfile1_")
+        assert host.path.exists(pathlib.Path("_testfile2_"))
+        # Case 2: Target file already exists.
+        #  Write the source file with a size different from the target file, so
+        #  we can check whether we find the old or the new stat information
+        #  when stat'ing the target file after the rename.
+        with host.open("_testfile1_", "w") as fobj:
+            fobj.write("abcdef\n")
+        self.make_remote_file("_testfile2_")
+        file1_stat = host.stat("_testfile1_")
+        file2_stat = host.stat("_testfile2_")
+        host.rename(pathlib.Path("_testfile1_"), "_testfile2_")
+        assert not host.path.exists("_testfile1_")
+        assert host.path.exists(pathlib.Path("_testfile2_"))
+        new_file2_stat = host.stat("_testfile2_")
+        assert new_file2_stat.st_size > file2_stat.st_size
+
+    def test_cache_invalidation_for_rename_exception(self):
+        """
+        If a file system item is renamed/moved, its stats information should be
+        removed from the cache. This should also work if the rename operation
+        raises an exception.
+        """
+        # Test for ticket #150
+        host = self.host
+        # Make sure the target of the renaming operation is removed later.
+        # Make sure both files are gone after the test.
+        self.cleaner.add_file("_testfile1_")
+        self.cleaner.add_file("_testfile2_")
+        # Write the source file with a size different from the target file, so
+        # we can check whether we find the old or the new stat information
+        # when stat'ing the target file after the rename.
+        with host.open("_testfile1_", "w") as fobj:
+            fobj.write("abcdef\n")
+        self.make_remote_file("_testfile2_")
+        file1_stat = host.stat("_testfile1_")
+        file2_stat = host.stat("_testfile2_")
+        # Monkey-patch session `rename` call.
+        old_rename = host._session.rename
+        def failing_rename(source, target):
+            # Simulate the case where the rename completely or partially
+            # succeeds on the server, but a proper reply doesn't get through to
+            # the client. It doesn't matter whether the exception is
+            # `error_temp` or `error_perm`.
+            old_rename(source, target)
+            raise ftplib.error_perm("simulated error")
+        host._session.rename = failing_rename
+        #
+        with pytest.raises(ftputil.error.PermanentError):
+            host.rename(pathlib.Path("_testfile1_"), "_testfile2_")
+        assert not host.path.exists("_testfile1_")
+        assert host.path.exists(pathlib.Path("_testfile2_"))
+        new_file2_stat = host.stat("_testfile2_")
+        assert new_file2_stat.st_size > file2_stat.st_size
+
     def test_rename_with_spaces_in_directory(self):
+        """
+        `rename` should work if source and target contain a directory with
+        spaces in its name.
+
+        This test also tests `Path` support.
+        """
         host = self.host
         dir_name = "_dir with spaces_"
         self.cleaner.add_dir(dir_name)
@@ -446,6 +672,9 @@ class TestRename(RealFTPTest):
 
 class TestStat(RealFTPTest):
     def test_stat(self):
+        """
+        Test some stat-related calls.
+        """
         host = self.host
         dir_name = "_testdir_"
         file_name = host.path.join(dir_name, "_nonempty_")
@@ -482,6 +711,10 @@ class TestStat(RealFTPTest):
         assert not (client_mtime - server_mtime > 120)
 
     def test_issomething_for_nonexistent_directory(self):
+        """
+        If the `is*` methods are called on a path where the base directory
+        doesn't exist, all of the methods should return `False`.
+        """
         host = self.host
         # Check if we get the right results if even the containing directory
         # doesn't exist (see ticket #66).
@@ -491,6 +724,11 @@ class TestStat(RealFTPTest):
         assert not host.path.islink(nonexistent_path)
 
     def test_special_broken_link(self):
+        """
+        If `is*` methods are called on a broken link (i.e. the link target
+        doesn't exist), `islink` should return `True` and `isdir` and `isfile`
+        should return `False`.
+        """
         # Test for ticket #39.
         host = self.host
         broken_link_name = os.path.join("dir_with_broken_link", "nonexistent")
@@ -500,6 +738,11 @@ class TestStat(RealFTPTest):
         assert host.path.islink(broken_link_name)
 
     def test_concurrent_access(self):
+        """
+        If we have two `FTPHost` instances for the same server directory,
+        operations on one `FTPHost` instance shouldn't influence the cache
+        entries of the other `FTPHost` instance.
+        """
         self.make_remote_file("_testfile_")
         with ftputil.FTPHost(
             *self.login_data, session_factory=DEFAULT_SESSION_FACTORY
@@ -553,6 +796,10 @@ class TestUploadAndDownload(RealFTPTest):
 
     @pytest.mark.slow_test
     def test_upload(self):
+        """
+        `upload_if_newer` should respect the time difference between client and
+        server, including consideration of the time shift.
+        """
         host = self.host
         host.synchronize_times()
         local_file = "_local_file_"
@@ -579,6 +826,10 @@ class TestUploadAndDownload(RealFTPTest):
 
     @pytest.mark.slow_test
     def test_download(self):
+        """
+        `download_if_newer` should respect the time difference between client
+        and server, including consideration of the time shift.
+        """
         host = self.host
         host.synchronize_times()
         local_file = "_local_file_"
@@ -612,6 +863,9 @@ class TestUploadAndDownload(RealFTPTest):
             os.unlink(local_file)
 
     def test_callback_with_transfer(self):
+        """
+        A `callback` argument should be called during a file transfer.
+        """
         host = self.host
         FILE_NAME = "large_file"
         # Default chunk size as in `FTPHost.copyfileobj`
@@ -645,6 +899,10 @@ class TestUploadAndDownload(RealFTPTest):
 
 class TestFTPFiles(RealFTPTest):
     def test_only_closed_children(self):
+        """
+        If `FTPHost.open` needs an `FTPFile` object, an existing closed
+        `FTPFile` should be reused.
+        """
         REMOTE_FILE_NAME = "CONTENTS"
         host = self.host
         with host.open(REMOTE_FILE_NAME, "rb") as file_obj1:
@@ -658,6 +916,9 @@ class TestFTPFiles(RealFTPTest):
                 assert file_obj._host is host._children[1]
 
     def test_no_timed_out_children(self):
+        """
+        `FTPHost.open` shouldn't use an `FTPFile` object that has timed out.
+        """
         REMOTE_FILE_NAME = "CONTENTS"
         host = self.host
         # Implicitly create child host object.
@@ -677,6 +938,10 @@ class TestFTPFiles(RealFTPTest):
         assert file_obj2 is file_obj3
 
     def test_no_delayed_226_children(self):
+        """
+        `FTPHost.open` shouldn't use an `FTPFile` object that gave an 226 error
+        reply.
+        """
         REMOTE_FILE_NAME = "CONTENTS"
         host = self.host
         # Implicitly create child host object.
@@ -733,6 +998,9 @@ class TestChmod(RealFTPTest):
         assert mode == expected_mode, "mode {0:o} != {1:o}".format(mode, expected_mode)
 
     def test_chmod_existing_directory(self):
+        """
+        A `chmod` operation on a writable existing directory should succeed.
+        """
         host = self.host
         host.mkdir("_test dir_")
         self.cleaner.add_dir("_test dir_")
@@ -746,6 +1014,9 @@ class TestChmod(RealFTPTest):
         self.assert_mode("_test dir_/nested_dir", 0o757)
 
     def test_chmod_existing_file(self):
+        """
+        A `chmod` operation on a writable existing file should succeed.
+        """
         host = self.host
         host.mkdir("_test dir_")
         self.cleaner.add_dir("_test dir_")
@@ -756,11 +1027,19 @@ class TestChmod(RealFTPTest):
         self.assert_mode(file_name, 0o646)
 
     def test_chmod_nonexistent_path(self):
+        """
+        If `chmod` is called on a non-existent path, a `PermanentError` should
+        be raised.
+        """
         # Set/get mode of a non-existing item.
         with pytest.raises(ftputil.error.PermanentError):
             self.host.chmod("nonexistent", 0o757)
 
     def test_cache_invalidation(self):
+        """
+        If the mode of a directory or file is changed, make sure that the cache
+        doesn't contain stat entries before the `chmod` call.
+        """
         host = self.host
         host.mkdir("_test dir_")
         self.cleaner.add_dir("_test dir_")
@@ -855,6 +1134,9 @@ class TestOther(RealFTPTest):
         file1.close()
 
     def test_subsequent_reading(self):
+        """
+        An `FTPHost.open` call should reuse a closed `FTPFile`.
+        """
         # Open a file for reading.
         with self.host.open("CONTENTS", "rb") as file1:
             pass
@@ -864,7 +1146,12 @@ class TestOther(RealFTPTest):
         assert file1._session is file2._session
 
     def test_names_with_spaces(self):
-        # Test if directories and files with spaces in their names can be used.
+        """
+        Directory and file names with spaces in them shouldn't cause problems.
+
+        (We've seen problems here with some servers before
+        `_robust_ftp_command` was introduced.)
+        """
         host = self.host
         assert host.path.isdir("dir with spaces")
         assert host.listdir("dir with spaces") == [
@@ -878,7 +1165,8 @@ class TestOther(RealFTPTest):
 
     def test_synchronize_times_without_write_access(self):
         """
-        Test failing synchronization because of non-writable directory.
+        If the current directory isn't writable, `synchronize_times` should
+        raise a `TimeShiftError`.
         """
         host = self.host
         # This isn't writable by the ftp account the tests are run under.
@@ -888,7 +1176,8 @@ class TestOther(RealFTPTest):
 
     def test_encoding(self):
         """
-        Test setting the path encoding.
+        If the session factory is requested with a specific encoding, the
+        session from the factory should use this requested encoding.
         """
         for encoding in ["latin-1", "UTF-8"]:
             factory = ftputil.session.session_factory(port=PORT, encoding=encoding)
@@ -941,6 +1230,16 @@ class TestOther(RealFTPTest):
             self.host.mkdir(path)
 
     def test_list_a_option(self):
+        """
+        If `use_list_a_option` is set to `False`, "hidden" files shouldn't be
+        included in the `listdir` result.
+
+        If `use_list_a_option` is set to `True`, "hidden" files should be
+        included in the `listdir` result.
+
+        Whether this works as described, depends on the FTP server and its
+        configuration.
+        """
         # For this test to pass, the server must _not_ list "hidden" files by
         # default but instead only when the `LIST` `-a` option is used.
         host = self.host
@@ -963,7 +1262,9 @@ class TestOther(RealFTPTest):
                         unused_data = fobj.read()
 
     def test_garbage_collection(self):
-        """Test whether there are cycles which prevent garbage collection."""
+        """
+        Test whether there are cycles which prevent garbage collection.
+        """
         gc.collect()
         objects_before_test = len(gc.garbage)
         self._make_objects_to_be_garbage_collected()
